{==============================================================================}
{ UNIT        BorderPanel  (can also be used as a button or a LED !)

  July 2010 - converted to LAZARUS (0.9.28.2) by BREAKOUTBOX
              FIXME:  FOCUS message not handled under LAZARUS
  June 2016 - changes applied to make BorderPanel run under Lauarus 1.6

  PROGRAMMER: Jürgen Sommer (original author)
  EMAIL:      webmaster@spiele-datenbank.de
  PURPOSE:    To enhance the normal TPanel with some graphical
              features like shaded corners and a new borderwidth
              property with which you can achieve some decent
              effects in your program.
  
  Please report bugs to webmaster@spiele-datenbank.de
  Every feedback is welcome :-)
                                                                               }
{==============================================================================}

unit BorderPanel;

interface

uses
  {$ifdef fpc}  // for LAZARUS
  Interfaces,
  LMessages, // contains the CM_xxx message definitions
  //JwaWindows,
  Windows,
  {$else}  // for Delphi
  Windows,
  Math,
  {$endif}
  Dialogs, Messages, SysUtils, Classes,
  Controls, ExtCtrls, Graphics, Forms;

type

  TBorderDirection = (bdUp, bdDown);
  TComponentMode = ( MPanel, MButton, MLed);

  TBorderPanel = class(TPanel)
  private
    { Private declarations }
    fColBorderFrame   : TColor;
    fColInnerFrame    : TColor;
    fColDownBackup    : TColor;
    fColFontHighlight : TColor;
    fColFontTemp      : TColor;
    //fColDefault       : TColor;   // new 2016-05-25 by Peter
    fColHighlight     : TColor;
    fColBackup        : TColor;
    fLEDOnColor       : TColor;
    fLEDOffColor      : TColor;
    fColFocusFrame    : TColor;
    fMode             : TComponentMode;
    fOnMouseEnter     : TNotifyEvent;
    fOnMouseLeave     : TNotifyEvent;
    fBorderDirection  : TBorderDirection;
    fMyBorderWidth    : integer;
    fBorderLight      : integer;
    fBorderDark       : integer;
    fActiveLED        : boolean;
    fInMouse          : Boolean;
    fActive           : boolean;
    fDefault          : boolean;
    fShowFocusFrame   : boolean;
    fShowFrame        : boolean;
    fShowInnerFrame   : boolean;
    function  GetBorderwidthNew: TBorderWidth;
    procedure SetBorderwidthNew(const Value: TBorderWidth);
    procedure SetFColBorderFrame(const Value: TColor);
    procedure SetFColInnerFrame(const Value: TColor);
    {HINT: this is the property to force TBorderPanel to behave as a button : }
    procedure SetFMode(const Value: TComponentMode);
    procedure SetLEDstatus(const Value: boolean);
    procedure SetFBorderdark(const Value: integer);
    procedure SetFBorderlight(const Value: integer);
    procedure SetFShowFrame(const Value: boolean);
    procedure SetFBorderDirection(const Value: TBorderdirection);
    procedure SetFShowInnerFrame(const Value: boolean);
    procedure SetLEDOnColor(Value: TColor);
    procedure SetLEDOffColor(Value: TColor);
    procedure WMLButtonDown(var aMessage: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonUp  (var aMessage: TWMLButtonUp);   message WM_LBUTTONUP;
    procedure CMMouseEnter (var aMessage: TMessage);       message CM_MOUSEENTER;
    procedure CMMouseLeave (var aMessage: TMessage);       message CM_MOUSELEAVE;
    procedure CMDialogKey  (var aMessage: TCMDialogKey);   message CM_DIALOGKEY{%H-};
    {$ifdef fpc}  // for LAZARUS
    // not jet implemented - see function code of "CMFocusChanged" !
    {$else}  // for Delphi
    procedure CMFocusChanged(var aMessage: TCMFocusChanged); message CM_FOCUSCHANGED;
    {$endif}
  protected
    { Protected declarations }
    procedure Click; override;
  public
    { Public declarations }
    constructor Create(AOwner:TComponent); override;
    procedure Paint; override;
  published
    { Published declarations }
    property Mode : TComponentMode
        read FMode write SetFMode default MPanel;
    property LEDOnColor : TColor
        read fLEDOnColor write SetLEDOnColor;
    property LEDOffColor : TColor
        read fLEDOffColor write SetLEDOffColor;
    property LEDstatus : boolean
        read fActiveLED write SetLEDstatus default TRUE;
    property ColFontHighlight : TColor
        read fColFontHighlight write FColFontHighlight;
    property ColBorderFrame : TColor
        read fColBorderFrame write SetFColBorderFrame;
    property BorderWidthNew : TBorderWidth
        read GetBorderwidthNew write SetBorderwidthNew default 4;
    property BorderFrameShow : boolean
        read fShowFrame write SetFShowFrame default true;
    property BorderInnerFrameShow : boolean
        read FShowInnerFrame write SetFShowInnerFrame default false;
    property ShowFocusFrame : boolean
        read fShowFocusFrame write FShowFocusFrame;
    property BorderLight : integer
        read fBorderlight write SetFBorderlight default 10;
    property BorderDark : integer
        read fBorderdark write SetFBorderdark default 10;
    property BorderDirection : TBorderdirection
        read fBorderDirection write SetFBorderDirection default bdUp;
    property OnMouseEnter: TNotifyEvent
        read fOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent
        read fOnMouseLeave write FOnMouseLeave;
    property ColInnerFrame : TColor
        read fColInnerFrame write SetFColInnerFrame;
    property ColHighlight : TColor
        read fColHighlight write FColHighlight;
    property ColFocusFrame: TColor
        read fColFocusFrame write FColFocusFrame;
  end;

{ the register procedure, mainly for Delphi : }
procedure Register;


implementation


{-------------------------- BRIGHTEN --------------------------}
function Brighten( Col:TColor; {%H-}Value:integer):TColor;
var
  rCol, gCol, bCol : integer;
begin
  col:= ColorToRGB( Col);
  rCol:= GetRvalue( Col);
  gCol:= GetGvalue( Col);
  bCol:= GetBvalue( Col);
  //if rcol < (254 -Value) then inc( rcol, value) else rcol:= 254;
  //if gcol < (254 -Value) then inc( gcol, value) else gcol:= 254;
  //if bcol < (254 -Value) then inc( bcol, value) else bcol:= 254;
  if rcol < 255-3 then inc( rCol, 3) else rcol:= 255;
  if gcol < 255-3 then inc( gCol, 3) else gcol:= 255;
  if bcol < 255-3 then inc( bCol, 3) else bcol:= 255;

  //Result:= RGB( rCol, gCol, bCol);
  Result:= RGBToColor( rCol, gCol, bCol);
end;

{-------------------------- DARKEN --------------------------}
function Darken(col:TColor; Value:dword; Steps:dword):TColor;
var
  rcol, gcol, bcol : dword;
  step : dword;
begin
  step:= max( Value +1, Steps div (Value+1));
  Col:= ColorToRGB( col);
  rcol:= GetRvalue( col);
  gcol:= GetGvalue( col);
  bcol:= GetBvalue( col);
  //if rcol > value then dec( rcol, value) else rcol:= 0;
  //if gcol > value then dec( gcol, value) else gcol:= 0;
  //if bcol > value then dec( bcol, value) else bcol:= 0;
  if rCol > step then dec( rcol, step) else rcol:= 0;
  if gCol > step then dec( gcol, step) else gcol:= 0;
  if bCol > step then dec( bcol, step) else bcol:= 0;

  //Result:= RGB( rCol, gCol, bCol);
  Result:= RGBToColor( rCol, gCol, bCol);
end;


{ ***** TBorderPanel ********************************************************* }
constructor TBorderPanel.Create(AOwner:TComponent);
begin
  inherited;

  { set default values : }
  Self.Color:=      clBtnFace;   // clDefault does not work here, whyever ..
  fColInnerFrame:=  clBlack;
  fColFocusFrame:=  clWhite;
  fColFontTemp:=    Font.Color;

  fBorderLight:=    9;
  fBorderDark:=     9;
  BevelOuter:=      bvNone;
  fActiveLED:=      TRUE;         // if TRUE then LEDcolor is = Self.Color
  if fLEDOnColor <> Color then fLEDOnColor:= Color;
  fDefault:=        FALSE;
  fShowFrame:=      FALSE;
  fMyBorderWidth:=  6;
  Caption:=         '';
  { Hint:  do NOT code
      BevelOuter:= bvRaised;
    because we paint frame ourselves ! }

  //fActive:= TRUE;  //used for DEBUG
  fActive:= FALSE;
end;

procedure TBorderPanel.Paint;
var
  i, k, n : integer;
  C : TColor;
begin
  inherited;

  with Canvas do
  begin
    { paint a Rect - only if component is active (HasFocus) and only in Delphi atm. }
    if (fMode = MButton) and (fActive) and (ShowFocusFrame)
      then
        begin
          Pen.Color:= fColFocusFrame;   // FixMe: darker color if Button is white !?
          Pen.Width:= 3;
          MoveTo( 0, 0);
          LineTo( Self.Width -1, 0);
          LineTo( Self.Width -1, Self.Height -1);
          LineTo( 0, Self.Height -1);
          LineTo( 0, 0);
          Pen.Width:= 1;   // reset pen width !
          k:= 1;
        end
      else k:= 0;  // HINT:  k is the offset to the border of the component
                   //        Only paint up to the border if NOT FActive !


  { paint an outer Rect if  BorderFrameShow = TRUE }
  if FShowFrame then     // HINT:
    if not FActive then    // do paint this Frame OR the FActive-Frame, not both !
      begin
        Pen.Color:= FColBorderFrame;
        MoveTo( 0, 0);
        LineTo( Self.Width-1, 0);
        LineTo( Self.Width-1, self.Height-1);
        LineTo( 0, Self.Height-1);
        LineTo( 0, 0);
        k:= 1;
      end;
      // .. else k:= 0;  is NOT reqired here as it has been set above !!


    if FShowInnerFrame then
      begin
        Pen.Color:= FColInnerFrame;
        MoveTo( FMyBorderWidth-1,           FMyBorderWidth -1);
        LineTo( Self.Width- fMyBorderWidth, FMyBorderWidth -1);
        LineTo( Self.Width- fMyBorderWidth, self.height -FMyBorderWidth);
        LineTo( FMyBorderWidth-1,           self.height -FMyBorderWidth);
        LineTo( FMyBorderWidth-1,           FMyBorderWidth -1);
        n:= 1;
      end
    else n:= 0;  // HINT:  n is the offset to the inner Frame
                 //        Only paint up to the inner Frame if NOT FActive !


    { ***** paint LEFT + TOP shading ***************************************** }
    C:= Self.Color;
    if fMyBorderWidth > 0 then
    for i:= fMyBorderWidth -1 -n downto k do
      begin
        if (FMode = mLed) or (fBorderDirection = bdDown)
          then C:= Darken( C, fBorderDark, fMyBorderWidth)
          else C:= Brighten( C, fBorderLight);
        Pen.Color:= C;
        MoveTo( i, Self.Height-i-k);
        LineTo( i, i);
        LineTo( Self.Width -i, i);
      end;


    { ***** paint RIGHT + BOTTOM shading ************************************* }
    C:= Self.Color;
    if FMyBorderWidth > 0 then
    for i:= FMyBorderWidth -1 -n downto k do
    begin
      if (FMode = mLed) or (fBorderDirection = bdUp)
        then C:= Darken( C, FBorderdark, FMyBorderWidth)
        else C:= Brighten( C, FBorderLight);
      Pen.Color:= C;
      MoveTo( Self.Width-i-k, i);
      LineTo( Self.Width-i-k, Self.Height-i-k);
      LineTo( i, Self.Height-i-k);
    end;

  end;  // with canvas do ..
end;


{ *** process messages ******************************************************* }
procedure TBorderPanel.WMLButtonDown (var aMessage: TWMLButtonDown);
begin
  inherited;

  if (fMode = MButton) and (BorderDirection <> bdDown) then
  begin
    fColDownBackup:= Self.Color;
    Self.Color:= Darken( Self.Color, 20, FMyBorderWidth);
    FBorderDirection:= bdDown;
    //Caption:= '$' +IntToHex( Self.Color, 8); // for DEBUG
    Repaint;
  end;
end;


procedure TBorderPanel.WMLButtonUp(var aMessage: TWMLButtonUp);
begin
  inherited;

  if (FMode = MButton) and (BorderDirection <> bdUp) then
    begin
      Self.Color:= fColDownBackup;
      FBorderDirection:= bdUp;
      if not fInMouse then Self.Color:= FColBackup;
      //Caption:= '$' +IntToHex( Self.Color, 8); // for DEBUG
      Repaint;
    end;
end;


procedure TBorderPanel.CMMouseEnter(var aMessage: TMessage);
begin
  inherited;

  fInMouse:= TRUE;
  if Assigned( fOnMouseEnter) then fOnMouseEnter( Self);

  if (fMode = MButton) then
    begin
      // backup colors :
      fColBackup:= Self.Color;
      FColFontTemp:= Font.Color;

      Self.Color:= FColHighlight;
      Font.Color:= fColFontHighlight;
      //Caption:= '$' +IntToHex( Self.Color, 8); // for DEBUG
      Repaint;
    end;
end;


procedure TBorderPanel.CMMouseLeave(var aMessage: TMessage);
begin
  inherited;

  fInMouse:= FALSE;
  if Assigned( fOnMouseLeave) then fOnMouseLeave( Self);

  if (fMode = MButton) and not (BorderDirection = bdDown) then
  begin
    // restore colors :
    Font.Color:= FColFontTemp;
    Self.Color:= FColBackup;

    //Caption:= '$' +IntToHex( Self.Color, 8); // for DEBUG
    Repaint;
  end;
end;

procedure TBorderPanel.CMDialogKey(var aMessage: TCMDialogKey);
begin
  with aMessage do
    if (((CharCode = VK_RETURN) and FActive) or
        ((CharCode = VK_SPACE) and FActive) or
        ((CharCode = VK_ESCAPE))) and
      (KeyDataToShiftState( aMessage.KeyData) = []) and CanFocus
    then
      begin
        if (FMode = MButton) then
          begin
            FColBackup:= Self.Color;
            Self.Color:= FColHighlight;
            Borderdirection:= bdDown;
            Repaint;
          end;
        Click;
        Result:= 1;
        if (FMode = MButton) then
          begin
            Self.Color:= FColBackup;
            Borderdirection:= bdUp;
            Repaint;
          end;
      end

    else inherited;
end;

{$ifdef fpc}  // for LAZARUS
// FixMe:  a platform independant implementation is required here
// Unfortunately FOCUS is handled different on several platforms
{$else}  // for Delphi
procedure TBorderPanel.CMFocusChanged(var aMessage: TCMFocusChanged);
begin
  with aMessage do
    if Sender is TBorderPanel
      then fActive:= Sender = Self
      else fActive:= FALSE;
    Repaint;
  inherited;
end;
{$endif}


procedure TBorderPanel.Click;   //FixMe:  whatfor ?   There's no additional code .. ?!
begin
  inherited;
end;

{ *** read properties ******************************************************** }
function TBorderPanel.GetBorderwidthNew: TBorderWidth;
begin
  Result:= FMyBorderWidth;
end;


{ *** write properties ******************************************************* }
procedure TBorderPanel.SetFBorderDark(const Value: integer);
begin
  fBorderDark:= Value;
  Repaint;
end;

procedure TBorderPanel.SetFBorderDirection(const Value: TBorderdirection);
begin
  if (FMode = MButton) then exit;
  FBorderDirection:= Value;
  Repaint;
end;

procedure TBorderPanel.SetFBorderlight(const Value: integer);
begin
  FBorderlight:= Value;
  Repaint;
end;

procedure TBorderPanel.SetFShowFrame(const Value: boolean);
begin
  FShowFrame:= Value;
  Repaint;
end;

procedure TBorderPanel.SetFShowInnerFrame(const Value: boolean);
begin
  FShowInnerFrame:= Value;
  Repaint;
end;

procedure TBorderPanel.SetBorderwidthNew(const Value: TBorderWidth);
var v : TBorderWidth;
    H, W : TBorderWidth;
begin
  { HINT: default value for FMyBorderWidth was :=  4 }
  H:= Self.Height;   // avoid converting operands to Int64
  W:= Self.Width;    // avoid converting operands to Int64

  v:= Value;
  { reduce BorderWidth Value to half of Height and Width : }
  if (v * 2 >= H) or (v * 2 >= W)
    then v:= min( H div 2, W div 2);
  FMyBorderWidth:= v;
  BorderWidth:= v;
end;

procedure TBorderPanel.SetFColBorderFrame(const Value: TColor);
begin
  FColBorderFrame:= Value;
  Repaint;
end;

procedure TBorderPanel.SetFColInnerFrame(const Value: TColor);
begin
  FColInnerFrame:= Value;
  Repaint;
end;

procedure TBorderPanel.SetFMode(const Value: TComponentMode);
begin
  if Value = MButton then
    begin
      FBorderDirection:= bdUp;
      Repaint;
    end;
	
  { avoid Parent Color : }				 
  if Value = MLed then ParentColor:= FALSE;

  FMode:= Value;
end;

procedure TBorderPanel.SetLEDOnColor(Value: TColor);
begin
  Self.fLEDOnColor:= Value;
end;

procedure TBorderPanel.SetLEDOffColor(Value: TColor);
begin
  Self.fLEDOffColor:= Value;
end;


procedure TBorderPanel.SetLEDstatus(const Value: boolean);
begin
  if fMode = MLed then
    begin
      fActiveLED:= Value;
      if fActiveLED
        then Self.Color:= Self.FLEDOnColor
        else Self.Color:= Self.FLEDOffColor;
    end;
end;


procedure Register;
begin
  { if integrated into a Delphi Package then drag it into a common 'Panels' tab : }
  RegisterComponents( 'Panels', [TBorderPanel]);
end;


end.
